import UniqueId
         
class eventmgr:
    def __init__(self):
        self.events={}
        
    def notify(self,name,msg=None,sendmsg=True):
        if name in self.events:
            for callback,arg,passarg in self.events[name]:
                if passarg:
                    if sendmsg:
                        callback(arg,msg)                
                    else:
                        callback(arg)
                else:
                    if sendmsg:
                        callback(msg)
                    else:
                        callback()
                    
    def bind(self,name,callback,arg=None,passarg=False,prepend=False):
        if not name in self.events:
            self.events[name]=[]
        if prepend:
            self.events[name].insert(0,(callback,arg,passarg))
        else:
            self.events[name].append((callback,arg,passarg))        
        
    def unbind(self,name,callback):
        i=-1
        eventlist=self.events[name]
        for n in xrange(len(eventlist)):
            if eventlist[n][0]==callback: 
                i=n
                break
        if i>=0:
            eventlist.pop(i)
            
    def free(self):
        self.events={}
        
        
        
        
idmgr=UniqueId.IDmgr()            
              
class baseclass(eventmgr):
    def __init__(self,mgr=None):
        eventmgr.__init__(self)
        self.uid=-1
        self.mgr=mgr
        if self.mgr<>None:
            self.mgr.addclass(self)
        
    def _attach(self,mgr):
        self.mgr=mgr
        self.idgroup='base-'+mgr.uids
        self.uid=idmgr.nextid(self.idgroup)
                
        self.notify('onattach',None,False)
    
    def _remove(self):
        if self.mgr<>None:
            idmgr.freeid(self.uid,self.idgroup)
            self.uid=-1
            self.mgr=None
            self.notify('onremove',None,False)            
            

    def free(self):
        self.notify('onfree',None,False)
        self.mgr.delobj(self.uid)        
        eventmgr.free(self)
        
    
class baseclass2(baseclass):
    def __init__(self):
        baseclass.__init__(self)
        self.attached={}
        self.unattached={}
        
        self.bind('onattach',self._attachevents)
        self.bind('onremove',self._detachevents)
        
    def addhandler(self,name,callback,takearg=True,arg=None,sendarg=False):
        assert(not name in self.unattached)
        assert(not name in self.attached)
        if sendarg:
            if takearg:            
                def cb(arg1,arg2=arg,callback=callback):
                    callback(arg1,arg2)
            else:
                def cb(arg2=arg,callback=callback):
                    callback(arg2)
            
        else:
            if takearg:            
                def cb(arg1,callback=callback):
                    callback(arg1)
            else:
                def cb(callback=callback):
                    callback()
                    
        if self.mgr<>None:            
            self.attached[name]=(cb)
            self.mgr.bind(name,cb)
        else:
            self.unattached[name]=(cb)
            
    def delhandler(self,name):
        if name in self.unattached:
            cb=self.unattached.pop(name)                    
            
        elif name in self.attached:
            cb=self.attached.pop(name)
                
        else: raise Exception("handler: "+name+" has not been added")
        
        if self.mgr<>None:
            self.mgr.unbind(name,cb)
            
    def _attachevents(self):
        for key,cb in self.unattached.iteritems():            
            self.mgr.bind(key,cb)
            self.attached[key]=cb
        self.unattached={}
            
    def _detachevents(self):
        for key,cb in self.attached.iteritems():
            self.mgr.unbind(key,cb)
            self.unattached[key]=cb
        self.attached={}
        
        
        
        
        
class basemgr(eventmgr):
    def __init__(self):
        eventmgr.__init__(self)
        self.uids=str(idmgr.nextid('basemgr'))
        self.objs={}
        
    def addobj(self,cls):
        cls._attach(self)
        self.objs[cls.uid]=cls
        self.notify('onaddobj',cls.uid)
        
    def delobj(self,uid):
        assert(uid in self.objs)
        self.notify('ondelobj',uid)
        obj=self.objs.pop(uid)
        obj._remove()
        
    def free(self):        
        self.notify('onfree',None,False)
        for uid in self.objs.keys():
            self.delobj(uid)
        eventmgr.free(self)
        
        
        
        
def test():
            
    c=baseclass()
    
    def onattach():
        print 'attached'
    
    def onremove():
        print 'removed'
    
    c.bind('onremove',onremove)
    c.bind('onattach',onattach)
    
    b=basemgr()
    b.addclass(c)
        
    print 'b uids:',b.uids
    print 'c uid:',c.uid
    
    b.free()
    
    
        
def test2():
            
    c=baseclass2()
    
    def onmsg(msg,cls):
        print 'cls:',cls.uid, 'msg:',msg
      
    c.addhandler('onmsg',onmsg,True,c,True)   
    
    b=basemgr()
    b.addclass(c)
        
    b.notify('onmsg','1-visible')

    c._remove()
    
    b.notify('onmsg','2-hidden')
    
    b.addclass(c)
    
    b.notify('onmsg','3-visible')
    
    c.delhandler('onmsg')
    
    b.notify('onmsg','4-hidden')
    
    c.addhandler('onmsg',onmsg,True,c,True)   
    
    b.notify('onmsg','5-visible')
    
    b.free()
    
    b.notify('onmsg','6')
    
if __name__=='__main__':
    test()
    print '---'
    test2()
        