from pandac.PandaModules import loadPrcFileData
loadPrcFileData("", """fullscreen 0
#show-frame-rate-meter 1
sync-video 0
background-color 0 0 0
win-size 800 600
interpolate-frames 1 #for animation
""")

FALLBACK_FULLSCREEN_SIZE=[1024,768]

import os#, sys
#import math

from eventmgr import eventmgr
from UniqueId import UniqueId 


from gui import *


import direct.directbase.DirectStart
base.disableMouse()


from pandac.PandaModules import *
from direct.showbase.DirectObject import DirectObject




class layer:
    def __init__(self,parentnp,name,sort):
        self.parentnp=parentnp
        self.name=name
        self.sort=sort
        self.np=parentnp.attachNewNode(name)
        self.np.setBin('fixed',sort)
        
        self.visible=True        

    def free(self):
        self.np.removeNode()
        self.np=None
        self.parentnp=None
        

    def show(self,visible):
        if visible<>self.visible:
            self.visible=visible
            if visible:
                self.np.show()
            else:
                self.np.hide()
   


class layermgr:
    def __init__(self,np):
        self.layers={}
        self.np=np
        
    def addlayer(self,name,sort=None):        
        assert(not name in self.layers)        
        if sort==None:
            sort=len(self.layers)
        self.layers[name]=layer(self.np,name,sort)
        
    def dellayer(self,name):
        l=self.layers.pop(name)
        assert(l<>None)
        l.free()

    def free(self):
        for l in self.layers.keys()[:]:
            self.dellayer(l)

        

class keybinder(feature,DirectObject):
    def __init__(self,parent,enabled=False):
        feature.__init__(self,parent)
        self.bind('onfree',self._onfree)
        self.keys={}
        self.enabled=enabled
        
    def setenabled(self,enabled=True):
        if enabled<>self.enabled:
            self.enabled=enabled            
            for key,down in self.keys.iteritems():
                if enabled:
                    self._enablekey(key)
                else:
                    if down:
                        self._onkey(key,False)
                    self._disablekey(key)            
    
    def iskeydown(self,key):
        return self.keys[key]
                    
    def _onkey(self,key,down):
        if self.keys[key]<>down:
            self.keys[key]=down                        
            e='onkey'+('up','down')[down]
            self.notify(e+'-'+key,None,False)
            
                                
    def _enablekey(self,key):
        self.accept(key,self._onkey,[key,True])
        self.accept(key+'-up',self._onkey,[key,False])
        
    def _disablekey(self,key):
        self.ignore(key)
        self.ignore(key+'-up')
                    
    def addkey(self,key):
        self.keys[key]=False
        if self.enabled:
            self._enablekey(key)
        
    def delkey(self,key):
        if self.enabled:
            self._disablekey(key)
            
    def _onfree(self):
        for key in self.keys.keys():
            self._disablekey(key)
#keybinder



class rootfeature(feature):
    def __init__(self):
        feature.__init__(self,None)              
        
        #self.TRACE=False
        
        #self.game=game                
        #self.game.addphystask('gamesim',1.0/80,self._updateroot)

        self.rootnp=render.attachNewNode('rootnp')
        self.rootnp.setDepthWrite(False)
        self.rootnp.setDepthTest(False)

        self.aspect2=aspect2d.attachNewNode('aspect2')
        
        self.layers={}   
        self.layermgrs={'render':layermgr(self.rootnp),
                        'aspect2d':layermgr(self.aspect2),
                        'hud':layermgr(render2d),     
                        }     

        self.gui=guibase(self)
         
    def free(self):        
        for l in self.layermgrs.values():
            l.free()
        self.rootnp.removeNode()
        self.rootnp=None
        self.aspect2.removeNode()
        self.aspect2=None
        feature.free(self)   
        
    def addlayer(self,group,name,sort=None):
        assert(not name in self.layers)
        self.layermgrs[group].addlayer(name,sort)
        self.layers[name]=self.layermgrs[group].layers[name].np


#    def _updateroot(self,dt):
#        self.gui._update(dt)
        
        
             
            
        



        
def settitle(title):
    wp=WindowProperties()
    wp.setTitle(title)
    base.win.requestProperties(wp)
        
def setfullscreen(full,size,pos):
    if full:
        if size[0]<10 or size[1]<10:
            size=FALLBACK_FULLSCREEN_SIZE
    wp=WindowProperties()
    wp.setFullscreen(full)    
    wp.setSize(size[0],size[1])
    if full:
        wp.setOrigin(0,0)
        wp.setUndecorated(True)
    else:
        wp.setOrigin(pos[0],pos[1])
        wp.setUndecorated(False)
    base.win.requestProperties(wp)
        
                

class gamebase(rootfeature):
    def __init__(self,title='Game',fps=80):        
        rootfeature.__init__(self)
        #eventmgr.__init__(self)
        
        settitle(title)
        
        self.time=0
        self.oldtime=0
        self.timers={}
    
        base.accept('escape',self.free)
        
        self.fullscreen=False
        self.oldsize=[0,0]
        base.accept('f11',self.togglefullscreen)
        
        self.addphystask('gamesim',1.0/fps,self._onsim)
        
        self.mousevisible=True
        self.centeredmouse=False
    
    def setcenteredmouse(self,centered):
        if self.centeredmouse<>centered:
            self.centeredmouse=centered
            
    def _recentermouse(self):        
        base.win.movePointer(0,base.win.getXSize()/2,base.win.getYSize()/2) 
        
    def togglemousevisible(self):
        self.setmousevisible(not self.mousevisible)
        
    def setmousevisible(self,visible):
        if visible<>self.mousevisible:
            self.mousevisible=visible
            
            props = WindowProperties() 
            props.setCursorHidden(not visible) 
            base.win.requestProperties(props)         
        
    def togglefullscreen(self):
        self.fullscreen=not self.fullscreen
        if self.fullscreen:
            self.oldsize=[base.win.getXSize(),base.win.getYSize()]
            wp=base.win.getProperties()
            self.oldpos=[wp.getXOrigin(),wp.getYOrigin()]
            setfullscreen(True,[base.pipe.getDisplayWidth(),base.pipe.getDisplayHeight()],[0,0])
        else:
            setfullscreen(False,self.oldsize,self.oldpos)
        
    def free(self):
        self.delphystask('gamesim')
        taskMgr.remove('run')
        rootfeature.free(self)
        import sys
        sys.exit()
        
    def run(self):
        taskMgr.add(self._ontask,'run')        
        run()
        
    def _onsim(self,dt):
        #self._updateroot(dt)        
        self.notify('onsim',dt)
        
    def _onframe(self,dt): 
        self.time=globalClock.getFrameTime()
        if self.centeredmouse:
            self._recentermouse()
        self._updatephystasks(dt)        
        self.notify('onframe',dt)
        
      
    def _ontask(self,task):
        self._onframe(globalClock.getDt())
        return task.cont
        
          
    #the callback may be called multiple times until it has caught up (dt==delay)    
    def addphystask(self,name,delay,callback):
        assert(delay>0)
        assert(not name in self.timers)                
        self.timers[name]=[delay,0,callback]   
    
    def delphystask(self,name):
        self.timers.pop(name)           
    
    def _updatephystasks(self,dt):
        for name,data in self.timers.copy().iteritems():            
            delay,elapsed,callback=data
            elapsed+=dt
            while elapsed>=delay:
                elapsed-=delay
                callback(delay)
            self.timers[name][1]=elapsed
        
        

 