from eventmgr import eventmgr
from pandac.PandaModules import *
import os

#import cPickle as pickle

from direct.gui.OnscreenText import OnscreenText

cardcache={}

CURSOR_NODRAG='nodrag32.png',22,[-1,11],False
CURSOR_DRADGDROP='dragdroparrow32.png',32,[3,28],True
CURSOR_FINGERPOINT='fingerpoint32.png',32,[7,3],True
CURSOR_NWSEARROWS='nwsearrows32.png',32,[15,15],True
CURSOR_NESWARROWS='neswarrows32.png',32,[15,15],True
CURSOR_EWARROWS='ewarrows32.png',32,[15,15],True
CURSOR_NSARROWS='nsarrows32.png',32,[15,15],True
CURSOR_NSEWARROWS='nsewarrows32.png',32,[15,15],True
CURSOR_POINTERSHADOW='pointershadow32-50.png',32,[3,3],False

UPDIRTEX='updir16.png'
GRIP32TEX='cornergrip32_g.png'


def appendmodelpath(dir):
    print 'adding modelpath: '+dir
    getModelPath().appendPath(dir)

appendmodelpath(os.path.join(os.getcwd(),'src','gui'))
appendmodelpath(os.path.join(os.getcwd(),'src','gui','borders'))





def copydict(d):
    r={}
    for key,val in d.iteritems():
        if isinstance(val,dict):
            r[key]=copydict(val)
        else:
            r[key]=val
    return r


cm=CardMaker('')    
def loadcard(nodepath=None,texture=None,scale=None,pos=None,ts=None,texscale=None,transparent=False,wrap=Texture.WMClamp,twosided=False,autorotate=True,cardframe=None,mipmap=False,centeruv=False):    

    if cardframe==None:    
        cardframe=[-0.5,0.5,-0.5,0.5]
        
    #cardhash=pickle.dumps([cardframe,centeruv])            
    cachedcard=None#cardcache.get(cardhash)
        
    if cachedcard==None:
        if centeruv:
            cm.setUvRange(Point2(-0.5,-0.5),Point2(0.5,0.5))
        else:
            cm.setUvRange(Point2(0,0),Point2(1,1))
        cm.setFrame(cardframe[0],cardframe[1],cardframe[2],cardframe[3])
        card=NodePath(cm.generate())
        #cardcache[cardhash]=card
        
    else:        
        #card=NodePath('')
        #cachedcard.instanceTo(card)
        card=cachedcard.instanceUnderNode(NodePath(''),'')
        
    if nodepath<>None:
        card.reparentTo(nodepath)
        
                                    
    if autorotate:
        card.setP(-90)
    
    if twosided:
        self.card.setTwoSided(True)
    
    if transparent:
        card.setTransparency(True)
        
    if scale<>None:
        card.setScale(scale)
        
    if pos<>None:
        card.setPos(pos)
    
    if texture<>None:
        if type(texture)==str:
            tx=loader.loadTexture(texture)
            if tx==None: 
                raise Exception('texture not found in loadcard:',texture)
        else:
            tx=texture            
        
        if texscale<>None:
            if ts<>None:
                card.setTexScale(ts,texscale[0],texscale[1])
            else:
                ts=TextureStage.getDefault()
                card.setTexScale(ts,texscale[0],texscale[1])
                
        if wrap<>None:
            if isinstance(wrap,(list,tuple)):
                tx.setWrapU(wrap[0])
                tx.setWrapV(wrap[1])
            else:
                tx.setWrapU(wrap)
                tx.setWrapV(wrap)
        
        if ts==None:                
            #ts=TextureStage('')
            card.setTexture(tx)
        else:            
            card.setTexture(ts,tx)
        
        if mipmap:                
            tx.setMinfilter(Texture.FTLinearMipmapLinear)
        
    return card


class feature(eventmgr):
    def __init__(self,parent):            
        eventmgr.__init__(self)        
        if parent==None:           
            self.root=self
        else:
            self.root=parent.root            
        self.parent=parent
        
    def free(self):                
        self.notify('onfree',None,False)
        eventmgr.free(self)
        self.parent=None





class widgetcommon(feature):
    def __init__(self,parent):    
        feature.__init__(self,parent)
        self.bind('onfree',self._onfree)
             
        
        self.children=[]
            
    def _addchild(self,child):
        self.children.append(child)
        child.bind('onsetpos',self._onchildmove)
        child.bind('onsetpos',self._onchildresize)
        self.childboundsdirty=True
        self.notify('onaddchild',child)
            
    def _delchild(self,child):
        self.children.remove(child)
        child.unbind('onsetpos',self._onchildmove)
        child.unbind('onsetpos',self._onchildresize)
        self.childboundsdirty=True
        self.notify('ondelchild',child)        
        
    def _notifychildren(self,msg,arg=None,sendarg=True):
        for child in self.children:
            child.notify(msg,arg,sendarg)
        
    def _onfree(self):
        for child in self.children:
            child.free()        


MNORMAL=0
MHOVER=1
MDOWN=2
MDRAG=3
MDOWNANDOUT=4
        
class guibase(widgetcommon):
    def __init__(self,parent):
        widgetcommon.__init__(self,parent)         
               
        self.bind('onfree',self._onfree)
        self.root.bind('onframe',self._update)
        
        self.windowsize=[0,0]
        self.mousepos=[0,0]

        self.guinp=render2d.attachNewNode('gui')
        self.guinp.setBin('fixed',-1)        
        self.guinp.setPos(-1,0,-1)
               
        self.mouse=self.guinp.attachNewNode('mouse')
        self.mouse.setBin('fixed',999)   
        
        self.cursoricon=loadcard(self.mouse,transparent=True,autorotate=False,cardframe=(0,1,0,1))                            
        #self.cursoricon.hide()        
        
      
        self.mousebuttons=[False,False,False]
        base.accept('mouse1',self._onmouse,[1,True])
        base.accept('mouse1-up',self._onmouse,[1,False])
        base.accept('mouse2',self._onmouse,[2,True])
        base.accept('mouse2-up',self._onmouse,[2,False])
        base.accept('mouse3',self._onmouse,[3,True])
        base.accept('mouse3-up',self._onmouse,[3,False])
                      
        self.mouseobj=None
        self.mousemode=MNORMAL
        self.mousedownpos=None
    
        self.cursormode='normal'
        self.cursortex=None
        
        self.children=[]
        
        self._update(0)
    
    def _setmouseobj(self,obj):
        if obj<>self.mouseobj:    
            self.mouseobj=obj
            self.notify('onsetmouseobj',obj)
            

    def _onfree(self):
        self.cursortex=None
        self._setmouseobj(None)        
        self.guinp.removeNode()
        self.mouse.removeNode()    
        base.ignore('mouse1')
        base.ignore('mouse1-up')
        base.ignore('mouse2')
        base.ignore('mouse2-up')
        base.ignore('mouse3')
        base.ignore('mouse3-up')

    def _addchild(self,child):
        self.children.append(child)
        self.notify('onaddchild',child)
        
    def _delchild(self,child):
        self.children.remove(child)
        self.notify('ondelchild',child)
        
    def ismousedown(self,button):
        return self.mousebuttons[button-1]
    
    def _hittest(self,pos):
        maxz=-1
        res=None
        for child in self.children:
            r=child._hittest(pos)
            if r<>None:
                if r.z>=maxz:
                    maxz=r.z
                    res=r        
        return res
    
    def _forcemouseup(self):
        if self.mousemode==MDRAG:
            self.mouseobj.notify('oncanceldrag',None,False)                       
                    
        elif self.mousemode in (MDOWN,MDOWNANDOUT):
            if self.mousebuttons[0]:
                button=1
            elif self.mousebuttons[1]:
                button=2
            elif self.mousebuttons[2]:
                button=3
            else:
                raise Exception('not possible')
            
            self.mouseobj.notify('onmouseup',button)
                    
        hr=self._hittest(self.mousepos)
        if hr<>None:
            self._setmouseobj(hr)
            self.mousemode=MHOVER
            hr.notify('onmouseenter',None,False)
        else:
            self._setmouseobj(None)        
            self.mousemode=MNORMAL
        self._setcursor('normal')        
            
    
    def _touchmouse(self):
        self._onmousemove(self.mousepos,self.mousepos)
        
    def _onwidgetenabled(self,sender):
        if sender==self.mouseobj or sender._ispointinbounds(self.mousepos):          
            #self._setcursor('normal')
            self._touchmouse()
        
    def _onwidgetdisabled(self,sender):
        if sender==self.mouseobj or sender._hittest(self.mousepos):
            self._setcursor('normal')
            self._touchmouse()
            
        
            
    def _setcursoricon(self,cursor):
        tx=None
        
        if cursor==None:
            cursor=CURSOR_POINTERSHADOW
                    
     
        assert(isinstance(cursor,(list,tuple)))
        assert(len(cursor[2])==2)
                
        tx=loader.loadTexture(cursor[0])
        if tx==None:
            raise Exception('cursor texture not found: '+cursor[0])            
            
        if tx<>self.cursortex:
            self.cursortex=tx            
            
            size=cursor[1]
            px,py=cursor[2]
            self.cursoricon.setScale(size) 
            self.cursoricon.setPos(-px,0,-size+py+1)
            self.cursoricon.setTexture(tx)            
            
            if cursor[3]:        
                self.root.setmousevisible(False)
            else:        
                self.root.setmousevisible(True)
        

    def _setcursor(self,mode):
        self.cursormode=mode 
            
        if mode=='normal':
            if self.mousemode in (MHOVER,MDOWN,MDRAG):
                if self.mouseobj.cursor<>None:                   
                    self._setcursoricon(self.mouseobj.cursor)
                else:
                    self._setcursoricon(None)
            else:
                self._setcursoricon(None)
                    
        elif mode=='nodrag':
            self._setcursoricon(CURSOR_NODRAG)
        elif mode=='dragdrop':
            self._setcursoricon(CURSOR_DRAGDROP)                

         
            
            
                 
    def _onmouse(self,button,down):
        downcount=0
        for i in xrange(3):
            if self.mousebuttons[i]: 
                downcount+=1
        
        self.mousebuttons[button-1]=down                        
        
        
        
        
        if down:                    
            if self.mousemode in (MDRAG,MDOWN,MDOWNANDOUT):
                self._forcemouseup()

                    
            elif self.mousemode==MHOVER and downcount==0:                    
                hr=self._hittest(self.mousepos)
                                
                if hr<>self.mouseobj:
                    self.mouseobj.notify('onmouseleave',None,False)
                    self._setmouseobj(hr)
                    if hr<>None:
                        self.mousemode=MHOVER       
                        hr.notify('onmouseenter',None,False)
                 
                        
                        
                if hr<>None:
                    self.mousemode=MDOWN  
                    if hr.draggable:    
                        self.mousedownpos=self.mousepos     
                    hr.notify('onmousedown',button)                
                else:
                    self.mousemode=MNORMAL
                         
                                                                
        #MOUSE UP
        else:
            hr=self._hittest(self.mousepos)
            
            if self.mousemode==MDRAG:
                    
                if hr<>None:
                    if hr<>self.mouseobj:
                    
                        r=hr._allowdragdrop(self.mouseobj)
                        if r:
                            self.mouseobj.notify('ondragdrop',hr)
                            hr.notify('onmouseenter',None,False)
                            self.mousemode=MHOVER                            
                            self._setmouseobj(hr)
                            #self._setcursor('normal') 
                        else:
                            self._forcemouseup()
                                                    
                    else:
                        self.mousemode=MHOVER                        
                        self.mouseobj.notify('oncanceldrag',None,False)                        
                        self.mouseobj.notify('onmouseenter',None,False)                                
                        #self._setcursor('normal')
                        
                else:
                    self._forcemouseup()      
                    
                self._setcursor('normal')              
                    
            elif self.mousemode in (MDOWN,MDOWNANDOUT):
                
                clicked=False
                if hr==self.mouseobj:                    
                    self.mousemode=MHOVER          
                    self.mouseobj.notify('onmouseup',button)
                    self._setcursor('normal')                          
                    hr.notify('onclick',None,False)
                    clicked=True
                else:                    
                    self.mousemode=MNORMAL #hack for button states
                    self.mouseobj.notify('onmouseup',button)
                    self.mouseobj.notify('onmouseleave',None,False)                  
                    
                if hr<>None:                        
                    self.mousemode=MHOVER
                    self._setmouseobj(hr)
                    hr.notify('onmouseenter',None,False)
                    
                else:
                    self.mousemode=MNORMAL
                    self._setmouseobj(None)
                    
                if not clicked:
                    self._setcursor('normal')
                  
        self.notify('onmouse'+str(button),down)
                   
        

    def _onmousemove(self,old,new):
        self.mousepos=new
        dx=new[0]-old[0]
        dy=new[1]-old[1]
        
        delta=[dx,dy]
        hr=self._hittest(new)
        
            
        if self.mousemode==MDRAG:        

            if hr<>None and hr<>self.mouseobj and self.mouseobj.allowdrop:                         
                dr=hr._allowdragdrop(self.mouseobj)
                if dr:
                    self.mouseobj.notify('ondragover',hr)
                    self._setcursor('dragdrop')
                else:
                    self.mouseobj.notify('ondrag',(new,delta))
                    self._setcursor('nodrag')
            else:            
                self.mouseobj.notify('ondrag',(new,delta))
                #self._setcursor('normal')
                
        elif self.mousemode in (MDOWN,MDOWNANDOUT):      
            
            if self.mouseobj.draggable:
                dragx=new[0]-self.mousedownpos[0]
                dragy=new[1]-self.mousedownpos[1]
                dd=self.mouseobj.dragdist
                startdrag=abs(dragx)>dd or abs(dragy)>dd
            else:
                startdrag=False
                
            if startdrag:
                if self.mousebuttons[0]:
                    button=1
                elif self.mousebuttons[1]:
                    button=2
                elif self.mousebuttons[2]:
                    button=3
                else:
                    raise Exception('not possible')
                
                self.mouseobj.notify('onmouseup',button)
                self.mousemode=MDRAG
                self.mouseobj.notify('onstartdrag',None,False)
                if self.mouseobj.allowdrop:
                    if hr<>self.mouseobj:
                        if hr<>None and hr.enabled:
                            if hr._allowdragdrop(self.mouseobj):
                                self._setcursor('dragdrop')
                            else:
                                self._setcursor('nodrag')
                        else:
                            self._setcursor('nodrag')
                    else:                        
                        self._setcursor('nodrag')            
                else:
                    self._setcursor('normal')
            else:
                if self.mousemode==MDOWNANDOUT:
                    if hr==self.mouseobj:
                        self.mousemode=MDOWN
                        hr.notify('onmouseenter',None,False)      
                        self._setcursor('normal')      
                
                else:
                    if self.mouseobj==hr:
                        hr.notify('onmousemove',(new,delta))                    
                    else:                    
                        self.mousemode=MDOWNANDOUT                    
                        self.mouseobj.notify('onmouseleave',None,False)
                        self._setcursor('normal')
                            
        elif self.mousemode==MHOVER:
            if hr<>None:
                if hr==self.mouseobj:
                    hr.notify('onmousemove',(new,delta))
                else:
                    self.mouseobj.notify('onmouseleave',None,False)                    
                    self._setmouseobj(hr)                    
                    hr.notify('onmouseenter',None,False)
                    self._setcursor('normal')
            else:
                self.mouseobj.notify('onmouseleave',None,False)
                self._setmouseobj(None)
                self.mousemode=MNORMAL
                self._setcursor('normal')
                    
                
        elif self.mousemode==MNORMAL:
            if hr<>None:
                self.mousemode=MHOVER
                self._setmouseobj(hr)
                hr.notify('onmouseenter',None,False)
                self._setcursor('normal')
    
        
    def _update(self,dt):
        wx=base.win.getXSize()
        wy=base.win.getYSize()
        if wy>0 and wx>0:            
            ws=[wx,wy]
            if ws<>self.windowsize:                
                self.windowsize=ws                
                self.guinp.setScale(2.0/wx,1,2.0/wy)                                                
                self._touchmouse()
                self.notify('onresize',ws)

        if base.mouseWatcherNode.hasMouse():
            self.mouse.show()
            mx=round((base.mouseWatcherNode.getMouseX()*0.5+0.5)*wx)
            my=round((base.mouseWatcherNode.getMouseY()*0.5+0.5)*wy-1)
            mp=[mx,my]
            if mp<>self.mousepos:                                               
                self.mouse.setPos(mx,0,my)                
                self._onmousemove(self.mousepos,mp)                
                self.notify('onmousemove',mp)
        else:
            self.mouse.hide()

        
        
class widget(widgetcommon):
    def __init__(self,parent,layout='nw',pos=[0,0],size=[300,300],cursor=None,draggable=False,visible=True,enabled=True,allowdrop=False,dragdist=3,zbias=0):
        widgetcommon.__init__(self,parent)
            
        self.ready=False
      
        #self.bind('onfree',self._onfree)
            
        self.wantsize=size
        self.wantpos=pos
        self.pos=[0,0]
        self.abspos=[0,0]
        self.size=[1,1]
                
        self.cursor=cursor
        self.allowdrop=allowdrop
        self.dragdist=dragdist        
        self.zbias=zbias
        self.z=0
        
        if type(layout)==str:
            self.layout=layoutclass(layout)
        elif layout==None:
            self.layout=layoutclass('nw')
        elif isinstance(layout,(list,tuple)):
            assert(len(layout)==2)
            self.layout=layoutclass('nw',layout)
        else:
            self.layout=layout
      
        self.visible=True
        self.wantvisible=visible
              
        self.enabled=True
        self.wantenabled=enabled

        
        self.draggable=draggable
        self.dragicon=None

        self.childbounds=None
        self.childboundsdirty=True
        
        self.bounds=None
        self.boundsdirty=True
        
        self.hidenullsize=False    

       # self.children=[]
            
        
    def setparent(self,parent):
        if parent<>self.parent:
            assert(parent<>None)            
            self.parent._delchild(self)
            self.parent=parent         
            parent._addchild()       
            self.notify('onchangeparent',parent)
    
    def _getbounds(self):
        if self.boundsdirty:
            x,y=self.abspos
            w,h=self.size
            b=[x,x+h,y,y+h]
            self.bounds=b
        else:
            return self.bounds
        
    def _getchildbounds(self):
        if self.childboundsdirty:
            self.childboundsdirty=False
            if len(self.children)>0:
                c=self.children[0]
                b=c._getbounds()                        
                for child in self.children[1:]:
                    cb=child._getbounds()                
                    if cb[0]<b[0]: b[0]=cb[0]        
                    if cb[1]>b[1]: b[1]=cb[1]        
                    if cb[2]<b[0]: b[2]=cb[2]        
                    if cb[3]>b[0]: b[3]=cb[3]        
                self.childbounds=b
                return b
            else:
                return [0,0,0,0]
        else:
            return self.childbounds
        
    def _hittest(self,pos):
        if not self.visible or not self.enabled: return None
        if not self._ispointinbounds(pos): return None
        maxz=self.z
        res=self
        for child in self.children:
            r=child._hittest(pos)
            if r<>None:
                if r.z>=maxz:
                    maxz=r.z
                    res=r
        return res

    def _onchildmove(self,posabspos):
        self.childboundsdirty=True
        
    def _onchildresize(self,size):
        self.childboundsdirty=True

    def _addchild(self,child):
        self.children.append(child)
        child.bind('onsetpos',self._onchildmove)
        child.bind('onsetpos',self._onchildresize)
        self.childboundsdirty=True
        self.notify('onaddchild',child)
            
    def _delchild(self,child):
        self.children.remove(child)
        child.unbind('onsetpos',self._onchildmove)
        child.unbind('onsetpos',self._onchildresize)
        self.childboundsdirty=True
        self.notify('ondelchild',child)
        
    def init(self):
        assert(not hasattr(self.layout,'parent') and not self.ready)
        self.layout.init(self)  
        self._doinit()      
        self.ready=True
        return self
    
    def _doinit(self): pass
    def _dosetz(self,z): pass      
    def _doresize(self,size): pass
    def _dosetpos(self,pos,abspos): pass
    def _dofree(self): pass
    def _dosetvisible(self,visible): pass
    def _dosetenabled(self,enabled): pass
        
    def setzbias(self,zbias):
        if self.zbias<>zbias:
            self.zbias=zbias
            self._setz(self.layout.parentz)
    
    def _setz(self,z):
        z+=self.zbias
        if z<>self.z:
            self.z=z
            self._dosetz(z)
            self.notify('onsetz',z)
            
        
    def _ispointinbounds(self,point):        
        x,y=self.abspos
        w,h=self.size
        
        if point[0]<x: return False
        if point[0]>=x+w: return False
        if point[1]<y: return False
        if point[1]>=y+h: return False
        return True
        
    def _allowdragdrop(self,sender):
        return False
        
    def _setenabled(self,enabled):
        enabled=enabled and self.layout.parentenabled
        if enabled<>self.enabled:
            self.enabled=enabled
            self._dosetenabled(enabled)                        
            if enabled:
                if self.visible:
                    self.root.gui._onwidgetenabled(self)
            else:
                if self.visible:
                    self.root.gui._onwidgetdisabled(self)
            self.notify('onsetenabled',enabled)
        
    def setenabled(self,enabled):
        if enabled<>self.wantenabled:
            self.wantenabled=enabled
            self._setenabled(enabled)
                
    def togglevisible(self):
        self.setvisible(not self.wantvisible)
          
    def _setvisible(self,visible):                 
        visible=visible and self.layout.parentvisible              
        if visible<>self.visible:
            
            visible=visible and not self.hidenullsize
            
            if visible<>self.visible:
                    
                self._dosetvisible(visible)                  
                if visible:
                    if self.enabled:
                        self.root.gui._onwidgetenabled(self)
                    
                else:
                    if self.enabled:
                        self.root.gui._onwidgetdisabled(self)
                    
            self.visible=visible
                    
        self.notify('onsetvisible',visible)                    
            
    def setvisible(self,visible):        
        if visible<>self.wantvisible:
            self.wantvisible=visible
            self._setvisible(visible)        
 
    def setsize(self,size):
        size=list(size)
        if size<>self.wantsize:
            self.wantsize=size
            self._setsize(size)
            
    def _setsize(self,size):        
        newsize=list(self.layout._calcsize(size))        
        if newsize<>self.size:     
            self.boundsdirty=True
            self.size=newsize       
            self._onresize()   
            self.notify('onresize',newsize)         
        else:
            self._setpos(self.wantpos)
            
    def _onresize(self):   
        
        if self.size[0]<1 or self.size[1]<1:                     
            if not self.hidenullsize:
                self.hidenullsize=True
                self._setvisible(False)            
        else:
            
            if self.hidenullsize:
                self.hidenullsize=False                                            
                self._setvisible(self.wantvisible and self.layout.parentvisible)              
               
            else:
                if self.visible and self.enabled:
                    self.root.gui._onwidgetenabled(self)        
                    
            self._setpos(self.wantpos)
            #self.notify('onsetpos',(self.pos,self.abspos))
            
            self._doresize(self.size)                    
              

            
    def _setpos(self,pos):
        localpos=self.layout._calcpos(pos)
        abx=localpos[0]+self.layout.parentpos[0]
        aby=localpos[1]+self.layout.parentpos[1]
        if localpos<>self.pos or abx<>self.abspos[0] or aby<>self.abspos[1]:
            self.boundsdirty=True
            self.pos=localpos
            self.abspos=[abx,aby]
            
            self._dosetpos(pos,self.abspos)            
            
            if self.visible and self.enabled:
                self.root.gui._onwidgetenabled(self)   
                
            self.notify('onsetpos',(pos,self.abspos))

            
    def setpos(self,pos):
        if pos<>self.wantpos:
            self.wantpos=pos            
            self._setpos(pos)
 
    def _onfree(self):
        self._dofree()
        widgetcommon._onfree(self)                
         



CORNERS={'ne':[1,1],'se':[1,0],'sw':[0,0],'nw':[0,1],'nc':[0.5,1],'sc':[0.5,0],'me':[1,0.5],'mw':[0,0.5],'mc':[0.5,0.5]}

class layoutclass:
    #if percentsize[0] is negative it uses -percentsize[0]*A where A=parentheight instead of parentwidth
    #if percentsize[1] is negative it uses -percentsize[1]*B where B=parentwidth instead of parentheight
    #the position of the widget adds an offset in the direction up,right except if cornered to N or E
    #if percentsize is used the widget's size is added afterwards
    def __init__(self,corner='nw',percentsize=None,getsizecallback=None):
        self.corner=corner        
        self.percentsize=percentsize
        self.getsizecallback=getsizecallback
        
        self.parentsize=[0,0]  
        self.parentpos=[0,0]
        self.parentz=0
        self.parentvisible=True
        self.parentenabled=True        
    
        self.widgetparent=False
                           
    def init(self,owner):
        self.owner=owner                
        owner.bind('onfree',self._free)
                
        if isinstance(owner.parent,widget):              
            self.widgetparent=True
            
            owner.parent.bind('onresize',self._onownerparentresize)
            owner.parent.bind('onsetpos',self._onownerparentsetpos)
            owner.parent.bind('onsetvisible',self._onownerparentsetvisible)
            owner.parent.bind('onsetz',self._onownerparentsetz)
            owner.parent.bind('onsetenabled',self._onownerparentsetenabled)                         
            
            owner.parent._addchild(owner)
            
            self._onownerparentsetz(owner.parent.z) 
            
            if owner.parent.ready: 
                self._onownerparentsetenabled(owner.parent.enabled)  
                self._onownerparentsetvisible(owner.parent.visible)                                
                self._onownerparentresize(owner.parent.size)
            else:
                self._onownerparentsetenabled(owner.parent.wantenabled)  
                self._onownerparentsetvisible(owner.parent.wantvisible)    
                
                
        else:            
            owner.root.gui.bind('onresize',self._onownerparentresize)         
               
            owner.root.gui._addchild(owner)
            
            self._onownerparentsetenabled(True)
            self._onownerparentsetvisible(True)
            self._onownerparentsetz(0)            
            self._onownerparentresize(owner.root.gui.windowsize)            
            
        
    def _onownerparentsetenabled(self,enabled):     
        self.parentenabled=enabled            
        self.owner._setenabled(self.owner.wantenabled)
        
            
    def _onownerparentsetvisible(self,visible):                
        self.parentvisible=visible
        self.owner._setvisible(self.owner.wantvisible)        
        
    def _onownerparentsetz(self,z):
        self.parentz=z
        self.owner._setz(z+1)
            
    def setpercentsize(self,percentsize):
        if percentsize<>self.percentsize:
            self.percentsize=percentsize
            self.owner._setsize(self.owner.wantsize)
            self.owner.notify('onsetpercentsize',percentsize)
            
    def setcorner(self,corner):    
        if corner<>self.corner:
            self.corner=corner 
            self.owner._setpos(self.owner.wantpos)             
            self.owner.notify('onsetcorner',corner)           
            
    def _calcpos(self,wantpos):
        size=self.owner.size
        pos=self.owner.wantpos        
        
        px,py=CORNERS[self.corner]
        x=self.parentsize[0]*px
        y=self.parentsize[1]*py
            
        if px==0.5:        
            x-=size[0]*0.5                                    
            x+=pos[0]
        else:
            if px==1:                
                x-=size[0]
                x-=pos[0]
            else:                                
                x+=pos[0]                 
                
        if py==0.5:     
            y-=size[1]*0.5    
            y+=pos[1]
        else:
            if py==1:
                y-=size[1]
                y-=pos[1]
            else:                                
                y+=pos[1]
                       
        return [round(x),round(y)]
            
    def _calcsize(self,wantsize):
        if self.getsizecallback<>None:
            return self.getsizecallback(wantsize)
        else:
            if self.percentsize<>None:                       
                px=self.percentsize[0]
                if px<0: 
                    px=abs(px)
                    pax=self.parentsize[1]
                else:
                    pax=self.parentsize[0]
                py=self.percentsize[1]
                if py<0:
                    py=abs(py)
                    pay=self.parentsize[0]
                else:
                    pay=self.parentsize[1]
                return [round(pax*px*0.01+wantsize[0]),round(pay*py*0.01+wantsize[1])]                                    
            else:            
                return wantsize
            
             
    def _onownerparentsetpos(self,posabspos):
        pos,abspos=posabspos
        self.parentpos=abspos
        self.owner._setpos(self.owner.wantpos)        
        self.owner.notify('onsetpos',(self.owner.pos,self.owner.abspos))
            
    def _onownerparentresize(self,ownersize):        
        self.parentsize=ownersize     
        self.owner._setsize(self.owner.wantsize)   
        self.owner.notify('onsetpos',(self.owner.pos,self.owner.abspos))
        
        
    def _free(self):   
        if self.widgetparent:
            self.owner.parent.unbind('onresize',self._onownerparentresize)
            self.owner.parent.unbind('onsetenabled',self._onownerparentsetenabled)            
            self.owner.parent.bind('onsetpos',self._onownerparentsetpos)
            self.owner.parent.bind('onsetvisible',self._onownerparentsetvisible)
            self.owner.parent.bind('onsetz',self._onownerparentsetz)
          
        else:            
            self.owner.root.gui.unbind('onresize',self._onownerparentresize)            
        self.owner.layout=None             
        self.owner=None
 
        
        
class label(widget):
    def __init__(self,parent,text,align='center',shadow=[0.03,0.03],shadowcolor=VBase4(0,0,0,1),wordwrap=True,fgcolor=VBase4(1,1,1,1),bgcolor=VBase4(0,0,0,0.5),widgetargs={}):
        widget.__init__(self,parent,**widgetargs)
      
        if align=='left':
            align=TextProperties.ALeft
        elif align=='right':
            align=TextProperties.ARight
        elif align=='center':
            align=TextProperties.ACenter
            
            
        self.align=align
        
        self.text=text
        self.wordwrap=wordwrap
        self.fgcolor=fgcolor
        self.bgcolor=bgcolor
        self.shadowcolor=shadowcolor
                
        self.textnode= TextNode('')
        self.textnode.setText(text)
        self.textnode.setAlign(align)
        #self.textnode.setBin('fixed')
        
        if shadow<>None:
            self.textnode.setShadow(shadow[0],shadow[1])
            self.textnode.setShadowColor(shadowcolor)                        
        
        self.textnode.setTextColor(fgcolor)
        self.textnode.setCardColor(bgcolor)
        self.label=self.root.gui.guinp.attachNewNode(self.textnode)
         
    def setfgcolor(self,fgcolor):
        if fgcolor<>self.fgcolor:
            self.fgcolor=fgcolor
            self.textnode.setTextColor(fgcolor)
    
    def setbgcolor(self,bgcolor):
        if bgcolor<>self.bgcolor:
            self.bgcolor=bgcolor
            self.textnode.setCardColor(bgcolor)     
            
    def setshadowcolor(self,shadowcolor):
        if shadowcolor<>self.shadowcolor:
            self.shadowcolor=shadowcolor
            self.textnode.setShadowColor(shadowcolor)
                    
    def _dofree(self):
        self.label.destroy() 
        self.label=None       
        self.textnode=None
        
    def appendtext(self,text):        
        self.textnode.appendText(text)
        
    def settext(self,text):        
        self.textnode.setText(text)
               
    def _dosetz(self,z):
        #self.textnode.setDrawOrder(z)
        self.label.setBin('fixed',z)
        
    def _dosetpos(self,pos,abspos):       
        self.label.setPos(abspos[0],0,abspos[1])    
        
    def _dosetvisible(self,visible):        
        if visible:
            self.label.show()
        else:            
            self.label.hide()
        
    def _doresize(self,size):
        self.label.setScale(size[0],1,size[1])        
        if self.wordwrap:
            self.textnode.setWordwrap(size[0])
        else:
            self.textnode.clearWordwrap()
        self.textnode.setCardActual(0,0,size[0],size[1])
               
        
        
        
class img(widget):
    #texscale: [a,b] -> [a,b]    
    #          True = [w/imgwid,h/imghei] taking intangle into account
    #          None = [1,1]  
    def __init__(self,parent,image=None,transparent=False,texscale=None,angle=0,intangle=0,wrap=Texture.WMClamp,mipmap=False,widgetargs={}):
       
        widget.__init__(self,parent,**widgetargs)         
                
        self.image=image
        #self.imagesize=imagesize
        self.transparent=transparent
        self.texscale=texscale
        self.angle=angle
        self.intangle=intangle
        
        self.ts=TextureStage('')     
             
        self.rotpath=self.root.gui.guinp.attachNewNode('')
        self.card=loadcard(nodepath=self.rotpath,texture=image,ts=self.ts,transparent=transparent,autorotate=False,wrap=wrap,mipmap=mipmap)       
        self.rotpath.setR(angle+intangle*90)                    
         
        if texscale<>None:
            if texscale<>True:
                self.settexscale(self.texscale)                
                
                
    def _dofree(self):    
        self.card.removeNode()  
        self.card=None
        self.rotpath.removeNode()      
        self.rotpath=None
        
    def _dosetvisible(self,visible):
        if visible:
            self.rotpath.show()
        else:
            self.rotpath.hide()
        
    def _dosetz(self,z):
        self.rotpath.setBin('fixed',z)       
        
    def _dosetpos(self,pos,abspos):
        self.rotpath.setPos(round(self.size[0]*0.5+abspos[0]),0,round(self.size[1]*0.5+abspos[1]-0.02))
        
    def _doresize(self,size):
        
        if self.intangle in (1,3):
            self.card.setScale(size[1],1,size[0])            
        else:
            self.card.setScale(size[0],1,size[1])    
            
        self.rotpath.setPos(round(size[0]*0.5+self.abspos[0]),0,round(size[1]*0.5+self.abspos[1]-0.02))
        
        self._rescaleimg()
       
    def _rescaleimg(self):
        #return
        if self.texscale==True:
            sx,sy=self.size       
            if sx<>0 and sy<>0:
                if sx>sy:
                    x=float(sx)/sy
                    y=1
                else:
                    x=1
                    y=float(sy)/sx
                    
                if self.intangle in (1,3):
                    self.card.setTexScale(self.ts,y,x,1)
                else:
                    self.card.setTexScale(self.ts,x,y,1)
        
        elif self.texscale==None:
            self.card.setTexScale(self.ts,1,1,1)            
        else:
            self.card.setTexScale(self.ts,self.texscale[0],self.texscale[1],1)            
                
    def setintangle(self,angle):        
        if angle<>self.intangle:
            self.intangle=angle
            self.rotpath.setR(self.angle+angle*90)            
            self._rescaleimg()
             
    def setangle(self,angle):        
        if angle<>self.angle:
            self.angle=angle
            self.rotpath.setR(angle+self.intangle*90)
                    
    def settexscale(self,texscale):
        if texscale<>self.texscale:
            self.texscale=texscale
            self._rescaleimg()            
        
    def setimage(self,img):
        if img<>None:
            if type(img)==str:
                img=loader.loadTexture(img)
                assert(img<>None)
                         
            if img<>self.image:
                self.image=img            
                self.card.setTexture(self.ts,self.image)
                        
         
                
                
#class widgetptr(widget):
#    def __init__(self,parent,ref=None,layout='mc',pos=(0,0),size=(300,300)):
#        widget.__init__(self,parent,layout,pos,size)        
    
        
#        self.ref=None
#        self.refpath=None
#        self.setref(ref)
        
#    def _dosetz(self,z):
#        self.nodepath.setBin('fized',z)
        
#    def _dofree(self):
#        self.ref=None
#        self.refpath.removeNode()
#        self.refpath=None
        
#    def setref(self,ref):        
#        if self.ref<>ref:        
                
#            if self.ref<>None:
#                self.ref.unbind('onfree',self._onreffree)
#                self.refpath.removeNode()
#                self.refpath=None
#                self.notify('onclearref',None,False)
#            if ref<>None:
#                self.refpath=self.root.gui.guinp.attachNewNode('')
#                ref.nodepath.instanceTo(self.refpath)
#                self.ref.bind('onfree',self._onreffree)
#                self.notify('onsetref',ref)        
#            self.ref=ref
            
#    def _onreffree(self):
#        self.refpath.removeNode()
#        self.refpath=None
#        self.ref=None
#        self.notify('onclearref',None,False)
        
#    def _onrefresize(self,size):
#        self.notify('onrefresize',size)        
   
        
BORDERS16=[('nw','corner16.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
           ('nc','top16.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]

BORDERS32=[('nw','corner32.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
           ('nc','top32.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]
           
BORDERSBLACK16=[('nw','corner16-black.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
           ('nc','top16-black.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]

BORDERSBLACK32=[('nw','corner32-black.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
           ('nc','top32-black.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]
        
BORDERSWHITE32=[('nw','whitecorner32.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
                ('nc','whitetop32.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]
                
BORDERSWHITE16=[('nw','whitecorner16.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
                ('nc','whitetop16.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]

BORDERSEMPTY16=[('nw','emptycorner16.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
                ('nc','emptytop16.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]
                
BORDERSEMPTY32=[('nw','emptycorner32.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
                ('nc','emptytop32.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]
                
BORDERSSHADOW32=[('nw','cornershadow32.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
                 ('nc','topshadow32.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]

BORDERSSHADOW64=[('nw','cornershadow64.png',0),('ne','nw',1),('sw','nw',3),('se','nw',2),
                 ('nc','topshadow64.png',0),('sc','nc',2),('me','nc',1),('mw','nc',3)]

BORDERNOBOTTOMSHADOW32=[('nw','cornershadow32.png',0),('ne','nw',1),('nc','topshadow32.png',0),('me','nc',1),('mw','nc',3)]

BORDERNOTOPSHADOW32=[('sw','cornershadow32.png',3),('se','sw',2),('sc','topshadow32.png',2),('me','sc',1),('mw','sc',3)]

class frame(widget):
    def __init__(self,parent,borders=BORDERS16,bordersize=16,bgcolor=VBase4(1.0,1.0,1.0,0.5),bordercolor=VBase4(1.0,1.0,1.0,0.5),clipped=False,widgetargs={}): 
                
        widget.__init__(self,parent,**widgetargs)            
          
        self.clipped=clipped
                     
        self.wantbordersize=0
        self.bordersize=0
        self.bgcolor=-1
        self.bordercolor=-1
        
        self.borderdata=borders
        self.borders={}            
             
        self.card=loadcard(self.root.gui.guinp,transparent=True,autorotate=False,cardframe=(0,1,0,1))                
       
        self.clipplanes=[]
        if self.clipped:        
            planes=[Vec3(1,0,0),Vec3(0,0,1),Vec3(-1,0,0),Vec3(0,0,-1)]
        
            for i in xrange(4):
                pl=Plane(planes[i],Point3(0,0,0))
                pn=PlaneNode('',pl)                       
                np=self.root.gui.guinp.attachNewNode(pn)   
                self.clipplanes.append(np)
                self.card.setClipPlane(np)
      
        self.edges={'me':[0,100],'mw':[0,100],'nc':[100,0],'sc':[100,0]}
                
        self.cardnullsized=False
                            
        self.setbgcolor(bgcolor)                                  
        self.setbordersize(bordersize)     
        self.setbordercolor(bordercolor)          
       
        
    
    def _dosetz(self,z):
        self.card.setBin('fixed',z)
        
    def _dosetvisible(self,visible):    
        if not self.cardnullsized:
            if visible:
                self.card.show()
            else:
                self.card.hide()
        
    def _doresize(self,size):
        sx=round(size[0]-self.bordersize*2)
        sy=round(size[1]-self.bordersize*2)
        
        
        if sx<1 or sy<1:
            if not self.cardnullsized:
                self.cardnullsized=True
                if self.visible:
                    self.card.hide()                
        else:
            if self.cardnullsized:
                self.cardnullsized=False
                if self.visible:
                    self.card.show()
                
            self.card.setScale(sx,1,sy)
            self._resetclip()
        
    def _resetclip(self):
        if self.clipped:
            sx=self.size[0]-self.bordersize*2
            sy=self.size[1]-self.bordersize*2        
            px=self.abspos[0]+self.bordersize
            py=self.abspos[1]+self.bordersize
           
            self.clipplanes[0].setPos(px,0,0)
            self.clipplanes[1].setPos(0,0,py)
            self.clipplanes[2].setPos(px+sx,0,0)
            self.clipplanes[3].setPos(0,0,py+sy)
   
            
    def _dosetpos(self,pos,abspos):
        self.card.setPos(round(abspos[0]+self.bordersize),0,round(abspos[1]+self.bordersize))    
        self._resetclip()
        
    def setbgcolor(self,bgcolor):
        if self.bgcolor<>bgcolor:
            self.bgcolor=bgcolor            
            self.card.setColor(bgcolor)
            
    def setbordercolor(self,bordercolor):        
        if self.bordercolor<>bordercolor:
            self.bordercolor=bordercolor
            if self.bordersize>0:
                for border in self.borders.values():
                    border.card.setColor(bordercolor)
        
    def _createborders(self):        
        corners=('ne','nw','se','sw')        
        
        for corner,image,ang in self.borderdata:                        

            size=[self.bordersize,self.bordersize]                                        
            wrap=[Texture.WMClamp,Texture.WMClamp]
            if corner in self.edges:                        
                percentscale=self.edges[corner]
                if percentscale[0]>0:
                    size[0]=-self.bordersize*2
                    wrap[1]=Texture.WMRepeat
                if percentscale[1]>0:
                    size[1]=-self.bordersize*2
                    wrap[0]=Texture.WMRepeat
            else:
                percentscale=[0,0]                        
                assert(corner in corners)
                
            if image in self.borders:                
                image=self.borders[image].image
            else:                
                assert(not (image in corners or image in self.edges))                        
                                              
            self.borders[corner]=img(self,image=image,widgetargs={'layout':layoutclass(corner,percentscale),'size':size,'pos':[0,0],'enabled':False},intangle=ang,transparent=True,texscale=None,wrap=wrap).init() #texscale=True
            
            
    def setbordersize(self,bordersize):
        if bordersize<>self.bordersize:
            self.bordersize=bordersize
            
            if bordersize<=0:
                for border in self.borders.values():
                    border.free()
                self.borders={}
            else:
                if self.borders=={}:
                    self._createborders()                          
           
                else:    
                    
                    for corner,border in self.borders.iteritems():
                        size=[bordersize,bordersize] 
                        if corner in self.edges:                                    
                            percentscale=self.edges[corner]
                            if percentscale[0]>0:
                                size[0]=-bordersize*2
                            if percentscale[1]>0:
                                size[1]=-bordersize*2
                        border.setsize(size)                        
                        
            self._dosetpos(self.pos,self.abspos)
            self.notify('onsetbordersize',bordersize)
                             
                
    def _dofree(self):
        self.setbordersize(0)
        self.card.removeNode()
        self.card=None
        
        
        

      
            
BNORMAL=0
BHOVER=1
BDOWN=2
BSELECTED=3
BDISABLED=4
        
class buttonbase(widget):
    #style options are: normal,radio,toggle
    #radiogroup can be None or a string which is the prefix of the buttondown event if the style is radio
    def __init__(self,parent,clickstyle='normal',pressed=False,widgetargs={},radiogroup=None):
        widget.__init__(self,parent,**widgetargs)    
        
        self.clickstyle=clickstyle
        
        self.bind('onenabled',self._onenabledchange)
        self.bind('ondisabled',self._onenabledchange)
        
        self.bind('onmouseup',self._setstate2,BNORMAL,True)
        self.bind('onmousedown',self._setstate2,BDOWN,True)
        self.bind('onmouseleave',self._setstate,BNORMAL,True)
        self.bind('onmouseenter',self._setstate,BHOVER,True)
     
        self.bind('onclick',self._onclick)
     
        self.radiogroup=None                
        self.pressed=False
        
        self.wantstate=BNORMAL
        self.state=BNORMAL
        
        self.setradiogroup(radiogroup)
        self.setpressed(pressed)
            
        #self._setstate(self.wantstate)
        
    def _dofree(self):
        self.setradiogroup(None)
        
    def setclickstyle(self,clickstyle):        
        if clickstyle<>self.clickstyle:
            assert(clickstyle in ('normal','radio','toggle'))
            self.clickstyle=clickstyle
            if clickstyle<>'normal':
                if self.pressed:
                    self._setstate(BDOWN)
                else:
                    self._setstate(BNORMAL)
        
    def setpressed(self,pressed):        
        if self.pressed<>pressed:
            self.pressed=pressed
            if self.clickstyle<>'normal':
                if pressed:
                    self._setstate(BDOWN)
                else:
                    self._setstate(BNORMAL)
                self.notify('onpressedchanged',pressed)
                
        
    def setradiogroup(self,radiogroup):
        if radiogroup<>self.radiogroup:
            if self.radiogroup<>None:
                self.unbind('radiogroup-'+self.radiogroup,self._onradiomessage)
            if radiogroup<>None:
                self.bind('radiogroup-'+radiogroup,self._onradiomessage)
            self.radiogroup=radiogroup
                    
    def _onradiomessage(self,msg):       
        msg,obj=msg
        if obj<>self:
            if msg=='onbuttonclick':                
                self.setpressed(False)            
        
    def _dosetstate(self,state): pass
        
    def _onenabledchange(self):
        self._setstate2(self.wantstate)    
        
    def _onclick(self):        
        if self.clickstyle<>'normal':                        
            if self.clickstyle=='radio':
                self.setpressed(True)  
                if self.radiogroup<>None:
                    if self.layout.widgetparent:
                        self.parent._notifychildren('radiogroup-'+self.radiogroup,('onbuttonclick',self))
                    else:            
                        self.root.gui._notifychildren('radiogroup-'+self.radiogroup,('onbuttonclick',self))
            else:
                self.setpressed(not self.pressed)  
                
    def _setstate2(self,state,button):                
        self._setstate(state)
        
    def _setstatefinal(self,state):            
        if self.state<>state:
            self.state=state
            self._dosetstate(state)
            self.notify('onstatechange',state)
            
    def _setstate(self,state):
        if not self.enabled:
            state=BDISABLED
            
        elif self.root.gui.mouseobj==self:
            if state==BHOVER and self.root.gui.mousemode==MDOWN:
                state=MDOWN
        
            elif state==BNORMAL:
                if self.root.gui.mousemode==MDOWNANDOUT:
                    if self.clickstyle<>'normal' and self.pressed:
                        state=BDOWN
                    else:
                        state=BNORMAL
                        
            if state==BNORMAL and self.clickstyle<>'normal' and self.pressed:            
                state=BDOWN
        else:            
            if self.clickstyle<>'normal' and self.pressed:            
                state=BDOWN
              
        if state<>self.wantstate:
            self.wantstate=state
            self._setstatefinal(state)
           
        
        


BUTTONSTYLE1={BNORMAL:{'fg':VBase4(0.1,0.4,0.8,1),'bg':VBase4(0.3,0.5,1,1),'sh':VBase4(0,0,0,1)},
               BHOVER:{'fg':VBase4(1,1,1,1),'bg':VBase4(0.3,0.6,1,1),'sh':VBase4(0,0,0,1)},
               BDOWN:{'fg':VBase4(0,0,0,0),'bg':VBase4(0.1,0.4,0.8,1),'sh':VBase4(1,1,1,1)},
               BSELECTED:{'fg':VBase4(1,1,1,1),'bg':VBase4(0.3,0.6,1,1)},
               BDISABLED:{'fg':VBase4(0.3,0.3,0.3,1),'bg':VBase4(0.7,0.6,0.5,1)},               
               }
        
                
class button(buttonbase):  
    def __init__(self,parent,text=None,image=None,clickstyle='normal',pressed=False,labelargs={},buttonstyle=BUTTONSTYLE1,radiogroup=None,frameargs={},widgetargs={}):                      
        if not 'cursor' in widgetargs:
            widgetargs['cursor']=CURSOR_FINGERPOINT
        buttonbase.__init__(self,parent,clickstyle=clickstyle,widgetargs=widgetargs,radiogroup=radiogroup,pressed=pressed)
        
        self.buttonstyle=buttonstyle
        
        if 'widgetargs' in labelargs:
            labelargs['widgetargs']['enabled']=False
            #labelargs['widgetargs']['zbias']=2
        else:
            labelargs['widgetargs']={'layout':[-80,80],'size':[0,0],'enabled':False}
           
        frameargs['widgetargs']={'layout':[100,100],'size':[0,0],'enabled':False,'pos':[0,0]}
        
             
        self.labelargs=labelargs               
            
        self.frame=frame(self,**frameargs).init()
        
        self.image=None
        self.img=None
        
        self.text=None
        self.label=None
                
        self.settext(text)
        self.setimage(image)
        
        self._dosetstate(self.state)
            
    def setimage(self,image):
        if image<>self.image:
            if self.img<>None:
                self.img.free()
                self.img=None
            self.image=image
            if image<>None:
                self.img=img(self,image=image,transparent=True).init()
        
    def _dosetstate(self,state):
        
        style=self.buttonstyle[state]
        self.frame.setbgcolor(style['bg'])
        self.frame.setbordercolor(style['bg'])
        if self.text<>None:
            self.label.setfgcolor(style['fg'])
            if 'sh' in style:
                self.label.setshadowcolor(style['sh'])
            else:
                self.label.setshadowcolor(Vec4(0,0,0,1))
         
    def setbuttonstyle(self,style):
        if style<>self.buttonstyle:
            self.buttonstyle=style
            self._dosetstate(self.state)
        
    def settext(self,text):
        if text<>self.text:
            self.text=text
            if text==None:
                if self.label<>None:
                    self.label.free()
                    self.label=None
                else:
                    raise Exception('how did we get here?')
            else:
                if self.label<>None:
                    self.label.settext(text)
                else:
                    self.label=label(self.frame,text=text,**self.labelargs).init()
                    self.label.settext(text)
                
        
    def _dofree(self):
        buttonbase._dofree(self)
        self.settext(None)
        self.setimage(None)
        
        
         
class window(frame):
    def __init__(self,parent,draggable=False,resizable=False,minsize=[50,50],maxsize=None,minimizebutton=False,minimized=False,keeponscreen=False,frameargs={}):
        frame.__init__(self,parent,**frameargs)
        
        self.keeponscreen=keeponscreen
        
        if keeponscreen:
            self.root.gui.bind('onresize',self._onscreenresize)        
     
        self.oldsize=None
        self.oldpos=None
              
                
        self.minsize=minsize
        self.maxsize=maxsize
        
               
        if draggable:
            
            self.bind('onmousedown',self._onmousedown)
            self.bind('ondrag',self._ontitledrag)

        self.draggable=draggable

        if resizable:
            self.grip=img(self,image=GRIP32TEX,widgetargs={'layout':'se','pos':[0,0],'draggable':True,'cursor':CURSOR_NWSEARROWS,'dragdist':0,'size':[self.bordersize,self.bordersize]},transparent=True).init()        
            self.grip.card.setColor(0,0,0,1)        
            self.grip.bind('onmousedown',self._ongripmousedown)
            self.grip.bind('ondrag',self._ondragcorner)
                

        if minimizebutton:
            bs=copydict(BUTTONSTYLE1)
            bs[BNORMAL]['bg']=Vec4(0,0,0,0)
            bs[BHOVER]['bg']=VBase4(1,1,0.5,0.5)
            bs[BDOWN]['bg']=VBase4(0.9,0.7,0.2,0.5)
        
            self.closebutton=button(self.parent,
                                    buttonstyle=bs,
                                    frameargs={'bordersize':self.bordersize*0.5,'borders':BORDERSWHITE16},
                                    widgetargs={'zbias':20,'size':[self.bordersize,self.bordersize],'layout':'sw'},                                
                                    ).init()
            self.closebutton.bind('onclick',self._togglevisible)
            self.closebutton.setpos([-1,-1])
        
            self.closeimg=img(self.closebutton,image=UPDIRTEX,widgetargs={'zbias':1,'layout':layoutclass('mc',[80,80]),'size':[0,0],'enabled':False},transparent=True).init()        
            self.closeimg.card.setColor(0,0,0,1)
            self.closeimg.setangle(-45)
            
            self.closeimg.setpos(self.abspos)
                            
            if minimized:
                self.closebutton.notify('onclick')
            
    def _doinit(self):
        self._ensureonscreen()
                
    def setminsize(self,minsize):
        if self.minsize<>minsize:
            self.minsize=minsize
            self._ensureonscreen()
            
    def setmaxsize(self,maxsize):
        if self.maxsize<>maxsize:
            self.maxsize=maxsize
            self._ensureonscreen()
        
    def _ensureonscreen(self):  
        if self.keeponscreen:      
            vx=self.wantsize[0]
            vy=self.wantsize[1]
            mx=self.root.gui.windowsize[0]
            my=self.root.gui.windowsize[1]
            vx=max(vx,self.minsize[0])
            vy=max(vy,self.minsize[1])        
            vx=min(vx,mx)
            vy=min(vy,my)                
            if self.maxsize<>None:
                vx=min(vx,self.maxsize[0])
                vy=min(vy,self.maxsize[1])                        
            self.setsize([vx,vy])
            
            vx=self.wantpos[0]
            vy=self.wantpos[1]
            vx=max(0,vx)                
            vy=max(0,vy)        
            mx=self.root.gui.windowsize[0]-self.wantsize[0]
            my=self.root.gui.windowsize[1]-self.wantsize[1]    
            vx=min(mx,vx)
            vy=min(my,vy)       
            self.setpos([vx,vy])
        
        
    def _ontitledrag(self,posdelta):        
        pos,delta=posdelta
        if delta<>[0,0]:
            xs=(1,-1)[self.layout.corner[1]=='e']
            ys=(-1,1)[self.layout.corner[0]=='n']
            
            vx=self.oldpos[0]+(pos[0]-self.root.gui.mousedownpos[0])*xs
            vx=max(0,vx)            
            vy=self.oldpos[1]-(pos[1]-self.root.gui.mousedownpos[1])*ys
            vy=max(0,vy)            
            mx=self.root.gui.windowsize[0]-self.wantsize[0]
            my=self.root.gui.windowsize[1]-self.wantsize[1]        
            vx=min(mx,vx)
            vy=min(my,vy)                                       
            self.setpos([vx,vy])
  
     
    def _onmousedown(self,button):
        self.oldpos=self.wantpos
                 
    def _ongripmousedown(self,button):
        self.oldsize=self.wantsize
            
    def _ondragcorner(self,posdelta):        
        pos,delta=posdelta
        if delta<>[0,0]:
            vx=self.oldsize[0]+pos[0]-self.root.gui.mousedownpos[0]                    
            vy=self.oldsize[1]-pos[1]+self.root.gui.mousedownpos[1]                        
            mx=self.root.gui.windowsize[0]#-self.wantpos[0]
            my=self.root.gui.windowsize[1]#-self.wantpos[1]        
            vx=max(vx,self.minsize[0])
            vy=max(vy,self.minsize[1])            
            vx=min(vx,mx)
            vy=min(vy,my)
            if self.maxsize<>None:
                vx=min(mx,self.maxsize[0])
                vy=min(my,self.maxsize[1])                             
                
            dx=vx-self.wantsize[0]
            dy=vy-self.wantsize[1]
            if dx<>0 or dy<>0:
                self.setsize([vx,vy])
            
                xs=self.layout.corner[1]<>'e'
                ys=self.layout.corner[0]=='n'                
                px=self.wantpos[0]
                if not xs:
                    px-=dx                
                py=self.wantpos[1]
                if not ys:
                    py-=dy                    
                self.setpos([px,py])
                
                self._ensureonscreen()
            
                    
    def _dosetpos(self,pos,abspos):        
        frame._dosetpos(self,pos,abspos)
        if hasattr(self,'closebutton'):
            
            y=abspos[1]+self.size[1]-self.bordersize
            
            mx=self.size[0]*0.5+abspos[0]
            
            if mx<=self.root.gui.windowsize[0]*0.5:
                x=abspos[0]
                self.closeimg.setangle(-45)                       
            else:
                x=abspos[0]+self.size[0]-self.bordersize
                self.closeimg.setangle(45)                            
                
            self.closebutton.setpos([x,y])
            
    def _togglevisible(self):
        self.togglevisible()
        self.closeimg.setintangle((self.closeimg.intangle+2)%4)
        
    def _onscreenresize(self,size):
        self._ensureonscreen()
        
        

class scrollbar(widget):
    def __init__(self,parent,direction='v',colors={'handleframe':VBase4(0,0,0,0),'handle':VBase4(0,0,0,0.1)},value=0.0,pagesize=0.01,range=[0.0,1.0],buttons=False,typ=float,widgetargs={}):
        widget.__init__(self,parent,**widgetargs)
        
        self.typ=typ                
        self.direction=direction
                    
        if direction=='v':
            flayout=layoutclass('nc',[100,0])            
            hlayout=layoutclass('nc',[100,0])
            blayout=layoutclass('nc',[100,-100])
            cursor=CURSOR_NSARROWS
        else:
            assert(direction=='h')
            flayout=layoutclass('mw',[0,100])            
            hlayout=layoutclass('mw',[0,100])
            blayout=layoutclass('mw',[-100,100])
            cursor=CURSOR_EWARROWS
                    
            
                            
        self.handleframe=frame(self,bordersize=0,bgcolor=colors['handleframe'],widgetargs={'layout':flayout,'size':[0,0]}).init()        
        self.handle=frame(self.handleframe,bordersize=0,bgcolor=colors['handle'],
                          widgetargs={'layout':hlayout,'size':[0,0],'draggable':True,'cursor':cursor},
                          ).init()
          
        self.handle.bind('ondrag',self._ondraghandle)        
        self.handle.bind('onmousedown',self._onhandlemousedown)
                    
        
        self.range=range
        self.value=value
        self.pagesize=pagesize
        self.handlesize=0.01
        self.handlepos=0
        
        self.dragoffdist=135
        
        self.scrollstartval=None
                  
        self._reconfig(value=value,pagesize=pagesize,range=range)
           
    
    def _onhandlemousedown(self,button):        
        self.scrollstartval=self.value
        
    def _doresize(self,size):
        if self.direction=='v':
            #self.handleframe.setsize([0,size[1]-size[0]*2])
            self.handleframe.setsize([0,size[1]])
            #self.handleframe.setpos([0,size[0]])
        else:
            #self.handleframe.setsize([size[0]-size[1]*2,0])
            self.handleframe.setsize([size[0],0])
            #self.handleframe.setpos([size[1],0])
        self._reconfig(range=self.range)
            
    def _ondraghandle(self,posdelta):
        pos,delta=posdelta
        axis,dir=([0,1],[1,-1])[self.direction=='v']
      
        d=(pos[axis]-self.root.gui.mousedownpos[axis])*dir   
        do=pos[1-axis]-self.root.gui.mousedownpos[1-axis]   
        if d<>0:
            if abs(do)>=self.dragoffdist:
                self.setvalue(self.scrollstartval)
            else:
                s=self.range[1]-self.range[0]
                r=s-self.pagesize
                siz=self.handleframe.size[self.direction=='v']
                d=d*s*s/(siz*r)
                self.setvalue(self.scrollstartval+d)
         
        
    def _calchandlesize(self):
        siz=self.size[self.direction=='v']
        return float(self.pagesize*100)/(self.range[1]-self.range[0])
        return s
    
    def _calchandlepos(self):
        s=self.range[1]-self.range[0]
        r=s-self.pagesize       
        v=self.value-self.range[0]                
        c=float(v)/s
        f=c*r/s
        siz=self.handleframe.size[self.direction=='v']
        p=round(f*siz)    
        return p
        
    def _reconfig(self,value=None,pagesize=None,range=None):            
        invervalchange=value<>None or pagesize<>None
        if invervalchange and range<>None:        
            self.handlesize=self._calchandlesize()            
        elif invervalchange:
            self.handlepos=self._calchandlepos()
        else:
            self.handlesize=self._calchandlesize()
            self.handlepos=self._calchandlepos()
        
        if self.direction=='v':
            self.handle.setpos([0,self.handlepos])            
            self.handle.layout.setpercentsize([100,self.handlesize])                                  
        else:
            self.handle.setpos([self.handlepos,0])                                
            self.handle.layout.setpercentsize([self.handlesize,100])        
            
            
    def addvalue(self,dvalue):
        self._setvalue(self.value+dvalue)            
            
    def _addvalue(self,dvalue):
        self._setvalue(self.value+dvalue)
            
    def setvalue(self,value):
        self._setvalue(value)
                    
    def setpagesize(self,size):
        self._setpagesize(size)        
            
    def _setpagesize(self,size):
        if size<>self.pagesize:
            self.pagesize=size
            self._reconfig(pagesize=size)
            self.notify('onpagesizechange',size)
            
    def _setvalue(self,value):
        value=self.typ(value)
        if value>self.range[1]:
            value=self.range[1]
        elif value<self.range[0]:
            value=self.range[0]            
        if value<>self.value:
            self.value=value
            self._reconfig(value=value)
            self.notify('onvaluechange',self.value)  

         
class scrolledframe(frame):
    def __init__(self,parent,frameargs={},subframeargs={},usescrollbars=[True,True]):
        frame.__init__(self,parent,**frameargs)                
        
        self.usescrollbars=usescrollbars
        self.alwaysusescrollbars=alwaysusescrollbars
        
        self.subframe=frame(self,**subframeargs)
        
        #if self.usescrollbars[0]:
        #    self.scrollbar1=scrollbar(self,
        
        self.subframebounds=[0,0,0,0]
            
    def _refitscrollbars(self):
        b=self.subframe._getchildbounds()
        if b<>self.subframebounds:            
            sx=b[1]-b[0]
            sy=b[3]-b[2]
            osx=self.subframebounds[1]-self.subframebounds[0]
            osy=self.subframebounds[3]-self.subframebounds[2]
            dx=sx-osx
            dy=sy-osy
            
            v1=self.scroll
            
            fx,fy=self.subframe.size
            rx=fx-sx            
            ry=fy-sy
            
            #if fx<0
            
            #if dx<>0:
                                
            self.subframebounds=b
            
        
            
    def _doresize(self,size):
        frame._doresize(self,size)
        
        self._refitscrollbars()
        
        if self.usescrollbars[0]:
            sx=size[0]-self.scrollbar1.size[0]
        else:
            sx=size[0]
            
        if self.usescrollbars[1]:
            sy=size[1]-self.scrollbar1.size[1]
        else:
            sy=size[1]
                
        self.subframe.setsize([sx,sy])
        
    

if __name__=='__main__':
    import os,sys    
    sys.path.append(os.path.dirname(os.getcwd()))
    os.chdir(os.path.dirname(os.getcwd()))
    
    import main
    main.start()
      
        