"""
                    trails.py    
    by
        ___ ____ ____ ____ ____ ____ ____ _  _ 
         |  |__/ |___ |___ |___ |  | |__/ |\/| 
         |  |  \ |___ |___ |    |__| |  \ |  | 
         
    updated by DRC (Nov 25, 2008)
                    
  
"""


from pandac.PandaModules import CardMaker, NodePath, Vec2, Vec3,Point3, Vec4, Point3
from pandac.PandaModules import TextureStage,GeomVertexReader,GeomVertexData, GeomVertexWriter, GeomTristrips, Geom, GeomNode, GeomVertexFormat

from direct.task.Task import Task

import os



class trailid:  
    def __init__(self,id,texture,size=1):        
        self.id=id
        
        self.texture=texture
        if type(texture)==str:
            self.texture=loader.loadTexture(texture)
        else:
            assert(self.texture)
               
        self.size=size
        self.isdirty=False
        self.headdirty=False            



           
def _getoriented(render,p1,p2,up,division,XVEC=Vec3(1,0,0)):                           
    #anp=render.attachNewNode('anp')
    anp=NodePath('anp')
    
    anp.setPos(p1) 
    anp.lookAt(p2,up) 
    
    if division<>0:
        anp.setR(anp,180*division/self.divisions)                        
    oriented=render.getRelativeVector(anp,XVEC)
    l=oriented.length()
            #self.render                                       
    #oriented.normalize()
    #anp.removeNode()                            
    return oriented/l,l


cm=CardMaker('trailscardmaker')
cm.setFrame(-1,1,0,1)
class TrailGenerator(object):
    
    
    def __init__(self,nodepath,nodedist=4,traillength=20,nodeangle=20,divisions=1,dir=(0,-1,0),texscale=1.0):
        self.nodedist=nodedist
        self.divisions=divisions
        self.trailLength=traillength        
        self.dir=dir
        self.texscale=texscale
        self.trails = {}
        self.heads={}
        self.trailsnp={}        
        self.first = True
        self.locked=False
        
        self.dColor = Vec4(1,1,1,0)
        self.cColor = Vec4(1,1,1,1)        
        
        self.dohead=True
        
        self.minangle=nodeangle
        
        self.render=nodepath.attachNewNode('trailsrender')
        
    def destroy(self):
        self.removeall()


    #trailid must provide an id, texture, and a size property
    def addtrail(self,trailid,point):
        if not self.locked:         
            assert(not trailid.id in self.trails)            
            self.trails[trailid.id]=([point],trailid)
            self.trailsnp[trailid.id]=None
            self.heads[trailid.id]=self._genhead(trailid)
        else:
            print "LOCKED IN addtrail"
        
    
    def removetrail(self,trailid):  
        if not self.locked:            
            h=self.heads.pop(trailid.id)
            if h<>None: h.removeNode()                
            m=self.trailsnp.pop(trailid.id)
            if m<>None: m.removeNode()
            self.trails.pop(trailid.id)
        else:
            print "LOCKED IN removetrail"
        

    def removeall(self):
        if not self.locked:           
            for pointlist,trail in self.trails.values():
                self.removetrail(trail)        
        else:
            print "LOCKED IN removeall"
             
    def appendpoint(self,trailid,point):       
        if not self.locked: 
            
            trail=self.trails[trailid.id][0]

            if len(trail)>2:
                if trailid.headdirty:
                    trail[0]=point
                else:                    
                    t=trail[1]
                    t2=trail[2]
                    p=Vec3(point[0],point[1],point[2])
                    op=Vec3(t[0],t[1],t[2])
                    op2=Vec3(t2[0],t2[1],t2[2])
                    ov=op-op2
                    ov.normalize()
                    nv=p-op
                    nv.normalize()
                    ang=ov.angleDeg(nv)
                    
                    dist=Vec3(point-t).length()            
                    #print 'trail',trailid.id,' dist:',dist
                    if dist>=self.nodedist and ang>=self.minangle:                    
                        
                        trail.insert(0,point)
                        trailid.isdirty=True            
                        return True
                    else:
                        trail[0]=point
                        trailid.headdirty=True            
                
            else:
                trail.insert(0,point)          
                trailid.isdirty=True 
                return True
        else: 
            print "LOCKED IN appendpoint"
        
    def insertpoint(self,trailid,pos,point):
        if not self.locked:
            trail=self.trails[trailid.id][0]
            if pos==0:
                self.appendpoint(trailid,point)
            else:
                assert(pos>=0 and pos<len(trail))
                trail.insert(pos,point)
                self.isditry=True
        else: 
            print "LOCKED IN insertpoint"
       
    def _genhead(self,trailid):
        tid=trailid.id
        
        ids=str(id(tid))
        head=self.render.attachNewNode('trails-head'+ids)          
        
        for division in xrange(self.divisions):
            divnode=head.attachNewNode('div-'+str(division))
            divnode.attachNewNode(cm.generate())
            
            
            divnode.setR(180*division/self.divisions)                        
            divnode.setP(divnode,-90)
   
            divnode.setDepthWrite(False)
            #divnode.setDepthTest(False)
            divnode.setTransparency(True)
            divnode.setTwoSided(True)        
            divnode.setTexture(trailid.texture)
            #divnode.setBin("transparent",0)
        return head

    def shrinktrail(self,trailid):
        if not self.locked:
            #points=self.trails[trailid.id][0]
            self.trails[trailid.id]=([],trailid)
            trailid.isdirty=True
        else: 
            print "LOCKED IN shrinktrail"

    def trimlastpoint(self,trailid):
        
        if not self.locked:
            points=self.trails[trailid.id][0]
            hp=(2,3)[self.dohead]
            if len(points)>=hp:
                points.pop(-1)
                if len(points)==hp-1:
                    if self.trailsnp[trailid.id]:
                        self.trailsnp[trailid.id].removeNode()
                        self.trailsnp[trailid.id]=None
                        self.heads[trailid.id].reparentTo(hidden)
                else:
                    trailid.isdirty=True
        else: 
            print "LOCKED IN trimlastpoint"

     
            
    def update(self):     
        self.locked=True
        
        for trailid,tt in self.trails.iteritems():
            trail,trailobj=tt
    
            x=(2,3)[self.dohead]
            if len(trail)>=x:
                ids=str(id(trailid))
                                                                             
                if trailobj.isdirty:
                                                        
                    #print "REGEN"
                    
                    while len(trail)>=self.trailLength:
                        trail.pop(-1)
                    
                    if self.trailsnp[trailid]:
                        self.trailsnp[trailid].removeNode()
                        
                    nn=self.render.attachNewNode('trailnp:'+ids)
                    
                    
                    nn.setTransparency(True)
                    nn.setTwoSided(True)        
                    #ts=TextureStage('')
                    #ts.setMode(TextureStage.MAdd)   
                    nn.setTexture(trailobj.texture)
                    #nn.setBin("transparent",0)
                    nn.setDepthWrite(False)
                    nn.setDepthTest(False)
                    #nn.setDepthTest(False)
                    self.trailsnp[trailid]=nn
                
                    #import random
                    #print "REGEN: ",random.randint(1,1000)

                    for division in xrange(self.divisions):
                        vdata=(GeomVertexData('vdata', GeomVertexFormat.getV3c4t2() , Geom.UHStatic))
                        vertex=(GeomVertexWriter(vdata, 'vertex'))
                        color=(GeomVertexWriter(vdata, 'color'))
                        uv=(GeomVertexWriter(vdata, 'texcoord'))
                        geom=Geom(vdata)                       
                        geomnode=GeomNode('')
                        geomnode.addGeom(geom)
                        
                        self.trailsnp[trailid].attachNewNode(geomnode)
                        
                        l=len(trail)-1
                        
                        cColor = Vec4(self.cColor[0],self.cColor[1],self.cColor[2],self.cColor[3])
                        dColor = (self.dColor-cColor)/l
                        
                        cUV = 0
                        #dUV = self.texscale/l
                        
                        if self.dohead:
                            #dUV+=cUV
                            cColor+=dColor

                        #if self.dohead:
                        #  cColor+=dColor
                        #  cUV+=dUV   
                        up=Vec3(self.dir[0],self.dir[1],self.dir[2])
                        
                       
                            

                        
                        start=x-1 
                        head = trail[start-1]
                        num=0
                        for link in trail[start:]:
                            
                            oriented,l2=_getoriented(self.render,head,link,up,division)                        
              
                            v=(oriented * trailobj.size)    
                            
                            cUV += self.texscale * l2
                            
                            if link==trail[start]:
                                v1 = head + v
                                v2 = head - v
                          
                                vertex.addData3f(v1); color.addData4f(cColor); uv.addData2f(0,0)
                                vertex.addData3f(v2); color.addData4f(cColor); uv.addData2f(1,0)
                                num+=2
              
                            v3 = link + v
                            v4 = link - v
                           

                            vertex.addData3f(v3); color.addData4f(cColor+dColor);  uv.addData2f(0,cUV)
                            vertex.addData3f(v4); color.addData4f(cColor+dColor);   uv.addData2f(1,cUV)
                   
                         
                            cColor += dColor
                            #dUVSmooth = dUV*(head-link).length()
                            
                            num+=2
                            head = link
                       
                        assert(num>0)
                        prim=GeomTristrips(Geom.UHStatic) 
                        prim.addConsecutiveVertices(0,num)
                        prim.closePrimitive()                    
                        geom.addPrimitive(prim)
                        
                           
                if self.dohead:  
                    if trailobj.isdirty or trailobj.headdirty:
                        trailobj.headdirty=False
                        
                        dist=Vec3(trail[0]-trail[1]).length()            
                                                   
                  
                        h=self.heads[trailid]
                        
                        if dist<0.001:
                            h.hide()                                
                        else:
                            h.show()                            
                            h.setPos(trail[0])
                            h.lookAt(trail[1],-Vec3(self.dir[0],self.dir[1],self.dir[2]))
                                                    
                            children=self.heads[trailid].getChildren()                                                        
                            for i in xrange(children.getNumPaths()):
                                headcard=children.getPath(i)
                                headcard.setTexScale(TextureStage.getDefault(),Vec2(1,self.texscale))                            
                                headcard.setScale(trailobj.size,trailobj.size,dist)

                                         
            else:
                if self.dohead:
                    head=self.heads[trailid]
                    if head: head.hide()
                    
                if trailid in self.trailsnp:
                    np=self.trailsnp[trailid]
                    if np: np.hide()
            
            trailobj.isdirty=False  
        self.locked=False
                       
       
